"""
    Temporal values and comparison
"""
from __future__ import annotations
import typing
import numpy as np
import numpy.typing as npt

__all__ = [
    "distant_past",
    "far_future",
    "isnever",
    "never"
]


def distant_past() -> float:
    """
    Returns a value that compares less than any other value but itself and "never"
    """
def far_future() -> float:
    """
    Returns a value that compares greater than any other value but itself and "never"
    """
@typing.overload
def isnever(t: float) -> bool:
    """
    Returns whether the value of t corresponds to "never". As "never" is implemented as a floating-point NaN,
    direct comparison will always fail, since NaN != NaN.
    """
@typing.overload
def isnever(t: npt.NDArray[np.float64] | list[float]) -> npt.NDArray[np.bool8]:
    """
    Returns an array of booleans corresponding to whether the element of an array correspond to "never". As "never" is
    implemented as a floating-point NaN, direct comparison will always fails, since NaN != NaN.
    """
def never() -> float:
    """
    Returns a value that compares unequal to any value, including itself.
    """



__version__ = "1.3.1"

from _neworder_core import (
  Model,
  MonteCarlo,
  Timeline,
  NoTimeline,
  LinearTimeline,
  NumericTimeline,
  CalendarTimeline,
  time, df, mpi, log, run, stats, checked, verbose
) # type: ignore
from .domain import Edge, Domain, Space, StateGrid
from .mc import as_np



"""
    Submodule for basic MPI environment discovery
"""
from __future__ import annotations

__all__ = [
    "rank",
    "size"
]


def rank() -> int:
    """
    Returns the MPI rank of the process
    """
def size() -> int:
    """
    Returns the MPI size (no. of processes) of the run
    """



import numpy as np
from _neworder_core import MonteCarlo  # type: ignore[import]


def as_np(mc: MonteCarlo) -> np.random.Generator:
  """
  Returns an adapter enabling the MonteCarlo object to be used with numpy random functionality
  """
  class _NpAdapter(np.random.BitGenerator):
    def __init__(self, rng: MonteCarlo):
      super().__init__(0)
      self.rng = rng
      self.rng.init_bitgen(self.capsule)  # type: ignore

  return np.random.Generator(_NpAdapter(mc)) # type: ignore



"""
A dynamic microsimulation framework";
"""
from __future__ import annotations
import typing
import datetime
import numpy as np
import numpy.typing as npt

import df  # type: ignore
import mpi # type: ignore
from . import time
import stats # type: ignore
from .domain import *

date_t = datetime.datetime | datetime.date
FloatArray1d = NPFloatArray | list[float]
NPIntArray = npt.NDArray[np.int64]
IntArray1d = NPIntArray | list[int]


__all__ = [
    "CalendarTimeline",
    "LinearTimeline",
    "Model",
    "MonteCarlo",
    "NoTimeline",
    "NumericTimeline",
    "Timeline",
    "checked",
    "df",
    "log",
    "mpi",
    "run",
    "stats",
    "time",
    "verbose"
]


class Timeline():
    def __init__(self) -> None: ...
    def __repr__(self) -> str:
        """
        Prints a human-readable representation of the timeline object
        """
    @property
    def at_end(self) -> bool:
        """
            Returns True if the current step is the end of the timeline

        :type: bool
        """
    @property
    def dt(self) -> float:
        """
            Returns the step size size of the timeline

        :type: float
        """
    @property
    def end(self) -> object:
        """
            Returns the time of the end of the timeline

        :type: object
        """
    @property
    def index(self) -> int:
        """
            Returns the index of the current step in the timeline

        :type: int
        """
    @property
    def nsteps(self) -> int:
        """
            Returns the number of steps in the timeline (or -1 if open-ended)

        :type: int
        """
    @property
    def start(self) -> object:
        """
            Returns the time of the start of the timeline

        :type: object
        """
    @property
    def time(self) -> object:
        """
            Returns the time of the current step in the timeline

        :type: object
        """
    pass
class LinearTimeline(Timeline):
    """
    An equally-spaced non-calendar timeline .
    """
    @typing.overload
    def __init__(self, start: float, end: float, nsteps: int) -> None:
        """
        Constructs a timeline from start to end, with the given number of steps.
        """
    @typing.overload
    def __init__(self, start: float, step: float) -> None:
        """
        Constructs an open-ended timeline give a start value and a step size. NB the model will run until the Model.halt() method is explicitly called
        (from inside the step() method). Note also that nsteps() will return -1 for timelines constructed this way
        """
        pass
class Model():
    """
    The base model class from which all neworder models should be subclassed
    """
    def __init__(self, timeline: Timeline, seeder: typing.Callable[[int], int] = MonteCarlo.deterministic_independent_stream) -> None:
        """
        Constructs a model object with a timeline and (optionally) a seeder function for the random stream(s)
        """
    def check(self) -> bool:
        """
        User-overridable method used to check internal state at each timestep.
        Default behaviour is to simply return True.
        Returning False will halt the model run.
        This function should not be called directly, it is used by the Model.run() function

        Returns:
            True if checks are ok, False otherwise.
        """
    def finalise(self) -> None:
        """
        User-overridable function for custom processing after the final step in the model run.
        Default behaviour does nothing. This function does not need to be called directly, it is called by the Model.run() function
        """
    def halt(self) -> None:
        """
        Signal to the model to stop execution gracefully at the end of the current timestep, e.g. if some convergence criterion has been met,
        or input is required from an upstream model. The model can be subsequently resumed by calling the run() function.
        For trapping exceptional/error conditions, prefer to raise an exception, or return False from the Model.check() function
        """
    def modify(self, r: int) -> None:
        """
        User-overridable method used to modify state in a per-process basis for multiprocess model runs.
        Default behaviour is to do nothing.
        This function should not be called directly, it is used by the Model.run() function
        """
    def step(self) -> None:
        """
        User-implemented method used to advance state of a model.
        Default behaviour raises NotImplementedError.
        This function should not be called directly, it is used by the Model.run() function
        """
    @property
    def mc(self) -> MonteCarlo:
        """
            The model's Monte-Carlo engine

        :type: MonteCarlo
        """
    @property
    def timeline(self) -> Timeline:
        """
            The model's timeline object

        :type: Timeline
        """
    pass
class MonteCarlo():
    """
    The model's Monte-Carlo engine with configurable options for parallel execution
    """
    def __repr__(self) -> str:
        """
        Prints a human-readable representation of the MonteCarlo engine
        """
    def arrivals(self, lambda_: FloatArray1d, dt: float, n: int, mingap: float) -> NPFloatArray:
        """
        Returns an array of n arrays of multiple arrival times from a nonhomogeneous Poisson process (with hazard rate lambda[i], time interval dt),
        with a minimum separation between events of mingap. Sampling uses the Lewis-Shedler "thinning" algorithm
        The final value of lambda must be zero, and thus arrivals don't always occur, indicated by a value of neworder.time.never()
        The inner dimension of the returned 2d array is governed by the the maximum number of arrivals sampled, and will thus vary
        """
    def counts(self, lambda_: FloatArray1d, dt: float) -> NPIntArray:
        """
        Returns an array of simulated arrival counts (within time dt) for each intensity in lambda
        """
    @staticmethod
    def deterministic_identical_stream(r: int) -> int:
        """
        Returns a deterministic seed (19937). Input argument is ignored
        """
    @staticmethod
    def deterministic_independent_stream(r: int) -> int:
        """
        Returns a deterministic seed that is a function of the input (19937+r).
        The model uses the MPI rank as the input argument, allowing for differently seeded streams in each process
        """
    @typing.overload
    def first_arrival(self, lambda_: FloatArray1d, dt: float, n: int) -> NPFloatArray:
        """
        Returns an array of length n of first arrival times from a nonhomogeneous Poisson process (with hazard rate lambda[i], time interval dt),
        with a minimum start time of minval. Sampling uses the Lewis-Shedler "thinning" algorithm
        If the final value of lambda is zero, no arrival is indicated by a value of neworder.time.never()
        """
    @typing.overload
    def first_arrival(self, lambda_: FloatArray1d, dt: float, n: int, minval: float) -> NPFloatArray:
        """
        Returns an array of length n of first arrival times from a nonhomogeneous Poisson process (with hazard rate lambda[i], time interval dt),
        with no minimum start time. Sampling uses the Lewis-Shedler "thinning" algorithm
        If the final value of lambda is zero, no arrival is indicated by a value of neworder.time.never()
        """
    @typing.overload
    def hazard(self, p: float, n: int) -> NPFloatArray:
        """
        Returns an array of ones (with hazard rate lambda) or zeros of length n
        """
    @typing.overload
    def hazard(self, p: NPFloatArray) -> NPFloatArray:
        """
        Returns an array of ones (with hazard rate lambda[i]) or zeros for each element in p
        """
    @typing.overload
    def next_arrival(self, startingpoints: FloatArray1d, lambda_: FloatArray1d, dt: float) -> NPFloatArray:
        """
        Returns an array of length n of subsequent arrival times from a nonhomogeneous Poisson process (with hazard rate lambda[i], time interval dt),
        with start times given by startingpoints with a minimum offset of mingap. Sampling uses the Lewis-Shedler "thinning" algorithm.
        If the relative flag is True, then lambda[0] corresponds to start time + mingap, not to absolute time
        If the final value of lambda is zero, no arrival is indicated by a value of neworder.time.never()
        """
    @typing.overload
    def next_arrival(self, startingpoints: FloatArray1d, lambda_: FloatArray1d, dt: float, relative: bool) -> NPFloatArray:
        """
        Returns an array of length n of subsequent arrival times from a nonhomogeneous Poisson process (with hazard rate lambda[i], time interval dt),
        with start times given by startingpoints. Sampling uses the Lewis-Shedler "thinning" algorithm.
        If the relative flag is True, then lambda[0] corresponds to start time, not to absolute time
        If the final value of lambda is zero, no arrival is indicated by a value of neworder.time.never()
        """
    @typing.overload
    def next_arrival(self, startingpoints: FloatArray1d, lambda_: FloatArray1d, dt: float, relative: bool, minsep: float) -> NPFloatArray:
        """
        Returns an array of length n of subsequent arrival times from a nonhomogeneous Poisson process (with hazard rate lambda[i], time interval dt),
        with start times given by startingpoints. Sampling uses the Lewis-Shedler "thinning" algorithm.
        If the final value of lambda is zero, no arrival is indicated by a value of neworder.time.never()
        """
    @staticmethod
    def nondeterministic_stream(r: int) -> int:
        """
        Returns a random seed from the platform's random_device. Input argument is ignored
        """
    def raw(self) -> int:
        """
        Returns a random 64-bit unsigned integer. Useful for seeding other generators.
        """
    def reset(self) -> None:
        """
        Resets the generator using the original seed.
        Use with care, esp in multi-process models with identical streams
        """
    def sample(self, n: int, cat_weights: NPFloatArray) -> NPIntArray:
        """
        Returns an array of length n containing randomly sampled categorical values, weighted according to cat_weights
        """
    def seed(self) -> int:
        """
        Returns the seed used to initialise the random stream
        """
    def state(self) -> int:
        """
        Returns a hash of the internal state of the generator. Avoids the extra complexity of tranmitting variable-length strings over MPI.
        """
    @typing.overload
    def stopping(self, lambda_: float, n: int) -> NPFloatArray:
        """
        Returns an array of stopping times (with hazard rate lambda) of length n
        """
    @typing.overload
    def stopping(self, lambda_: NPFloatArray) -> NPFloatArray:
        """
        Returns an array of stopping times (with hazard rate lambda[i]) for each element in lambda
        """
    def ustream(self, n: int) -> NPFloatArray:
        """
        Returns an array of uniform random [0,1) variates of length n
        """
    pass
class NoTimeline(Timeline):
    """
    An arbitrary one step timeline, for continuous-time models with no explicit (discrete) timeline
    """
    def __init__(self) -> None:
        """
        Constructs an arbitrary one step timeline, where the start and end times are undefined and there is a single step of size zero. Useful for continuous-time models
        """
    pass
class NumericTimeline(Timeline):
    """
    An custom non-calendar timeline where the user explicitly specifies the time points, which must be monotonically increasing.
    """
    def __init__(self, times: typing.List[float]) -> None:
        """
        Constructs a timeline from an array of time points.
        """
    pass
class CalendarTimeline(Timeline):
    """
    A calendar-based timeline
    """
    @typing.overload
    def __init__(self, start: date_t, end: date_t, step: int, unit: str) -> None:
        """
        Constructs a calendar-based timeline, given start and end dates, an increment specified as a multiple of days, months or years
        """
    @typing.overload
    def __init__(self, start: date_t, step: int, unit: str) -> None:
        """
        Constructs an open-ended calendar-based timeline, given a start date and an increment specified as a multiple of days, months or years.
         NB the model will run until the Model.halt() method is explicitly called (from inside the step() method). Note also that nsteps() will
         return -1 for timelines constructed this way
        """
    pass
def checked(checked: bool = True) -> None:
    """
    Sets the checked flag, which determines whether the model runs checks during execution
    """
def log(obj: object) -> None:
    """
    The logging function. Prints obj to the console, annotated with process information
    """
def run(model: Model) -> bool:
    """
    Runs the model. If the model has previously run it will resume from the point at which it was given the "halt" instruction. This is useful
    for external processing of model data, and/or feedback from external sources. If the model has already reached the end of the timeline, this
    function will have no effect. To re-run the model from the start, you must construct a new model object.
    Returns:
        True if model succeeded, False otherwise
    """
def verbose(verbose: bool = True) -> None:
    """
    Sets the verbose flag, which toggles detailed runtime logs
    """

def as_np(mc: MonteCarlo) -> np.random.Generator:
  """
  Returns an adapter enabling the MonteCarlo object to be used with numpy random functionality
  """



"""
Spatial structures for positioning and moving entities and computing distances
"""

from __future__ import annotations
from typing import Union, Optional, Any, Callable, TypeVar
from enum import Enum, auto
import numpy as np
import numpy.typing as npt
from scipy import signal  # type: ignore

NPFloatArray = npt.NDArray[np.float64]

class Edge(Enum):
  """
  Edge behaviour
  """
  UNBOUNDED = auto()
  WRAP = auto()
  CONSTRAIN = auto()
  BOUNCE = auto()


class Domain:
  """
  Base class for spatial domains.
  """

  def __init__(self, dim: int, edge: Edge, continuous: bool):
    self.__dim = dim
    self.__edge = edge
    self.__continuous = continuous

  @property
  def dim(self) -> int:
    """ The dimension of the spatial domain """
    return self.__dim

  @property
  def edge(self) -> Edge:
    """ The tyoe of edge constraint """
    return self.__edge

  @property
  def continuous(self) -> bool:
    """ Whether space is continuous or discrete """
    return self.__continuous


class Space(Domain):
  """
  Continuous rectangular n-dimensional finite or infinite domain.
  If finite, positioning and/or movement near the domain boundary is
  dictated by the `wrap` attribute.
  """

  @staticmethod
  def unbounded(dim: int) -> Space:
    """ Construct an unbounded Space """
    assert dim > 0
    return Space(np.full(dim, -np.inf), np.full(dim, +np.inf), edge=Edge.UNBOUNDED)

  def __init__(self, min: NPFloatArray, max: NPFloatArray, edge: Edge=Edge.CONSTRAIN):
    assert len(min) and len(min) == len(max)
    super().__init__(len(min), edge, True)

    # Space supports all edge behaviours
    assert edge in [Edge.UNBOUNDED, Edge.WRAP, Edge.CONSTRAIN, Edge.BOUNCE]

    assert np.all(max > min)

    self.min = min
    self.max = max

  @property
  def extent(self) -> tuple[NPFloatArray, NPFloatArray]:
    """ The extent of the space in terms of two opposing points """
    return self.min, self.max

  def move(self, positions: Any,
                 velocities: Any,
                 delta_t: float,
                 ungroup: bool=False) -> tuple[NPFloatArray, NPFloatArray]:
    """ Returns translated positions AND velocities """
    # group tuples into a single array if necessary
    if isinstance(positions, tuple):
      positions = np.column_stack(positions)
    if isinstance(velocities, tuple):
      velocities = np.column_stack(velocities)

    assert positions.dtype == float
    assert velocities.dtype == float
    assert positions.shape[-1] == self.dim and velocities.shape[-1] == self.dim
    if self.edge == Edge.UNBOUNDED:
      p = positions + velocities * delta_t
      v = velocities
    elif self.edge == Edge.CONSTRAIN:
      p = positions + velocities * delta_t
      v = velocities
      hitmin = np.where(p < self.min, 1, 0)
      p = np.where(hitmin, self.min, p)
      v = np.where(hitmin, 0, v)
      hitmax = np.where(p > self.max, 1, 0)
      p = np.where(hitmax, self.max, p)
      v = np.where(hitmax, 0, v)
    elif self.edge == Edge.BOUNCE:
      p = positions + velocities * delta_t
      v = velocities
      hitmin = np.where(p < self.min, 1, 0)
      p = np.where(hitmin, 2 * self.min - p, p)
      v = np.where(hitmin, -v, v)
      hitmax = np.where(p > self.max, 1, 0)
      p = np.where(hitmax, 2 * self.max - p, p)
      v = np.where(hitmax, -v, v)
    else:
      p = self.min + np.mod(positions + velocities * delta_t - self.min, self.max - self.min)
      v = velocities

    if ungroup:
      p = np.split(p, self.dim, axis=1)
      v = np.split(v, self.dim, axis=1)
    return p, v

  def dists2(self, positions: Union[tuple[NPFloatArray, ...], NPFloatArray], to_points: Optional[NPFloatArray]=None) -> tuple[NPFloatArray, Any]:
    """ The squared distance between points and separations along each axis """
    # group tuples into a single array if necessary
    if isinstance(positions, tuple):
      positions = np.column_stack(positions)
    if isinstance(to_points, tuple):
      to_points = np.column_stack(to_points)
    # distances w.r.t. self if to_points not explicitly specified
    if to_points is None:
      to_points = positions
    assert positions.dtype == float
    assert to_points.dtype == float
    n = positions.shape[0]
    m = to_points.shape[0]
    d = positions.shape[1]
    d2 = np.zeros((m, n))
    separations: tuple[Any, ...] = ()
    if self.edge != Edge.WRAP:
      for i in range(d):
        delta = np.tile(positions[:, i], m).reshape((m, n)) - np.repeat(to_points[:, i], n).reshape(m, n)
        separations += (delta,)
        d2 += delta * delta
    else: # wrapped domains need special treatment - distance across an edge may be less than naive distance
      for i in range(d):
        delta = np.tile(positions[:, i], m).reshape((m, n)) - np.repeat(to_points[:, i], n).reshape(m, n)
        r = self.max[i] - self.min[i]
        delta = np.where(delta > r / 2, delta - r, delta)
        delta = np.where(delta < -r / 2, delta + r, delta)
        separations += (delta,)
        d2 += delta * delta

    return d2, separations

  def dists(self, positions: Union[tuple[NPFloatArray, ...], NPFloatArray], to_points: Optional[NPFloatArray]=None) -> NPFloatArray:
    """ Returns distances between the points"""
    return np.sqrt(self.dists2(positions, to_points)[0])

  def in_range(self, distance: Any, positions: Any, count: Optional[bool]=False) -> NPFloatArray:
    """ Returns either indices or counts of points within the specified distance from each point """
    ind = np.where(self.dists2(positions)[0] < distance * distance, 1, 0)
    # fill diagonal so as not to include self - TODO how does this work if to_points!=positions
    np.fill_diagonal(ind, 0)
    return ind if not count else np.sum(ind, axis=1)

  def __repr__(self) -> str:
    return "%s dim=%d min=%s max=%s edge=%s" % (self.__class__.__name__, self.dim, self.min, self.max, self.edge)

def _bounce(i: int, N: int) -> int:
  s = (i // N) % 2
  k = i % N
  return N * s + (-1) ** s * k

class StateGrid(Domain):
  """
  Discrete rectangular n-dimensional finite grid domain with each cell having an integer state.
  Allows for counting of neighbours according to the supported edge behaviours:
    CONSTRAIN (no neighburs over edge), WRAP (toroidal), BOUNCE (reflect)
  """

  __mode_lookup: dict[Edge, str] = {
    Edge.CONSTRAIN: "constant",
    Edge.WRAP: "wrap",
    Edge.BOUNCE: "reflect"
  }

  def __init__(self, initial_values: NPFloatArray, edge: Edge=Edge.CONSTRAIN):
    super().__init__(initial_values.ndim, edge, False)

    # StateGrid supports two edge behaviours
    if edge not in [Edge.WRAP, Edge.CONSTRAIN, Edge.BOUNCE]:
      raise ValueError("edge policy must be one of Edge.WRAP, Edge.CONSTRAIN, Edge.BOUNCE")

    if initial_values.ndim < 1:
      raise ValueError("state array must have dimension of 1 or above")
    if initial_values.size < 1:
      raise ValueError("state array must have size of 1 or above in every dimension")

    self.state = initial_values

    # int neighbour kernel (not including self)
    self.kernel = np.ones([3] * self.dim)
    self.kernel[(1,) * self.dim] = 0

  def __get_point(self, p: tuple[int, ...]) -> tuple[int, ...]:
    assert len(p) == self.state.ndim, f"dimensionality mismatch: {len(p)} but grid has {self.state.ndim}"
    match self.edge:
      case Edge.WRAP:
        p = tuple(p[i] % self.state.shape[i] for i in range(len(p)))
      case Edge.CONSTRAIN:
        p = tuple(np.clip(p[i], 0, self.state.shape[i] - 1) for i in range(len(p)))
      case Edge.BOUNCE:
        p = tuple(_bounce(p[i], self.state.shape[i] - 1) for i in range(len(p)))
    return p

  def __getitem__(self, p: tuple[int, ...]) -> Any:
    return self.state[self.__get_point(p)]

  def __setitem__(self, p: tuple[int, ...], value: Any) -> None:
    self.state[self.__get_point(p)] = value

  def shift(self, position: tuple[int, ...], delta: tuple[int, ...]) -> tuple[int, ...]:
    """This translates a point according to the grid's edge behaviour. It does *not* change any state"""
    point = self.__get_point(position)
    p = tuple(point[i] + delta[i] for i in range(self.dim))
    return self.__get_point(p)

  @property
  def extent(self) -> Any:
    """ The extent of the space in terms of two opposing points """
    return self.state.shape

  def count_neighbours(self, indicator: Callable[[float], bool]=lambda x: x == 1) -> NPFloatArray:
    """ Counts neighbouring cells with a state indicated by supplied indicator function """

    ind: NPFloatArray = np.array([indicator(x) for x in self.state]).astype(int)  # automagically preserves shape
    # pad with boundary according to edge policy
    bounded = np.pad(ind, pad_width=1, mode=self.__mode_lookup[self.edge])  # type: ignore # bug?

    # count neighbours, drop padding, covert to int
    count = signal.convolve(bounded, self.kernel, mode="same", method="direct")[(slice(1, -1),) * self.dim].astype(int)

    return count




"""
    Submodule for operations involving direct manipulation of pandas dataframes
"""
from __future__ import annotations
from typing import TypeVar
import neworder
import numpy as np
# _Shape = typing.Tuple[int, ...]

T = TypeVar("T")
nparray = np.ndarray[T, np.dtype[T]]

__all__ = [
    "testfunc",
    "transition",
    "unique_index"
]


def testfunc(model: neworder.Model, df: object, colname: str) -> None:
    """
    Test function for direct dataframe manipulation. Results may vary. Do not use.
    """
def transition(model: neworder.Model, categories: nparray[np.int64], transition_matrix: nparray[np.float64], df: object, colname: str) -> None:
    """
    Randomly changes categorical data in a dataframe, according to supplied transition probabilities.
    Args:
        model: The model (for access to the MonteCarlo engine).
        categories: The set of possible categories
        transition_matrix: The probabilities of transitions between categories
        df: The dataframe, which is modified in-place
        colname: The name of the column in the dataframe
    """
def unique_index(n: int) -> nparray[np.int64]:
    """
    Generates an array of n unique values, even across multiple processes, that can be used to unambiguously index multiple dataframes.
    """



"""
    Submodule for statistical functions
"""
from __future__ import annotations
from typing import overload, TypeVar
import numpy as np
#_Shape = typing.Tuple[int, ...]

__all__ = [
    "logistic",
    "logit"
]

T = TypeVar("T")
nparray = np.ndarray[T, np.dtype[T]]

@overload
def logistic(x: nparray[np.float64]) -> nparray[np.float64]:
    """
    Computes the logistic function on the supplied values.
    Args:
        x: The input values.
        k: The growth rate
        x0: the midpoint location
    Returns:
        The function values


    Computes the logistic function with x0=0 on the supplied values.
    Args:
        x: The input values.
        k: The growth rate
    Returns:
        The function values


    Computes the logistic function with k=1 and x0=0 on the supplied values.
    Args:
        x: The input values.
    Returns:
        The function values
    """
@overload
def logistic(x: nparray[np.float64], k: float) -> nparray[np.float64]:
    pass
@overload
def logistic(x: nparray[np.float64], x0: float, k: float) -> nparray[np.float64]:
    pass
def logit(x: nparray[np.float64]) -> nparray[np.float64]:
    """
    Computes the logit function on the supplied values.
    Args:
        x: The input probability values in (0,1).
    Returns:
        The function values (log-odds)
    """



#!/usr/bin/env python3

import os
import glob
from setuptools import setup
import numpy
from pybind11.setup_helpers import Pybind11Extension, ParallelCompile

# see https://github.com/pybind/python_example


def readme():
  with open('README.md') as f:
    return f.read()


def list_files(dirs, exts, exclude=[]):
  files = []
  if isinstance(exclude, str):
    exclude = [exclude]
  for directory in dirs:
    for ext in exts:
      files.extend(glob.glob(os.path.join(directory, "*." + ext)))
  [f in files and files.remove(f) for f in exclude]
  return files


ext_modules = [
  Pybind11Extension(
    '_neworder_core',
    sources=list_files(['src'], ["cpp"]),
    include_dirs=[numpy.get_include()],
    depends=["setup.py", "neworder/__init__.py"] + list_files(["src"], ["h"]),
    cxx_std=20,
  ),
]

ParallelCompile().install()

setup(
  name='neworder',
  # version set from __init__.py via setup.cfg,
  author='Andrew P Smith',
  author_email='andrew@friarswood.net',
  url='https://neworder.readthedocs.io',
  description='A dynamic microsimulation framework',
  long_description=readme(),
  long_description_content_type="text/markdown",
  packages=["neworder"],
  package_data={"neworder": ["py.typed", "*.pyi"]},
  ext_modules=ext_modules,
  install_requires=['pandas>=1.0.5', 'pandas-stubs', 'scipy'],
  setup_requires=['pybind11>=2.5.0', 'pytest-runner', 'numpy>=1.19.1'],
  tests_require=['pytest', 'mpi4py>=3.0.3'],
  classifiers=[
    "Programming Language :: Python :: 3",
    "License :: OSI Approved :: MIT License",
    "Operating System :: OS Independent",
  ],
  zip_safe=False,
)
